"""
Functions for transferring landmarks from one mesh to another.

This module requires entente to be installed with the `landmarker` extras:

    pip install entente[landmarker]
"""

from cached_property import cached_property


class Landmarker(object):
    """
    An object which encapsulates a source mesh and a set of landmarks on that
    mesh. Its function is to transfer those landmarks onto a new mesh.

    The resultant landmarks will always be on or near the surface of the mesh.

    Args:
        source_mesh (lacecore.Mesh): The source mesh
        landmarks (dict): A mapping of landmark names to the points, which are
            `3x1` arraylike objects.
    """

    def __init__(self, source_mesh, landmarks):
        if not source_mesh.is_tri:
            raise ValueError("Source mesh should be triangulated")

        self.source_mesh = source_mesh
        self.landmarks = landmarks

    @classmethod
    def load(cls, source_mesh_path, landmark_path):
        """
        Create a landmarker using the given paths to a source mesh and landmarks.

        Args:
            source_mesh_path (str): File path to the source mesh.
            landmark_path (str): File path to a meshlab ``.pp`` file containing
                the landmark points.
        """
        import lacecore
        import meshlab_pickedpoints

        return cls(
            source_mesh=lacecore.load_obj(source_mesh_path, triangulate=True),
            landmarks=meshlab_pickedpoints.load(landmark_path),
        )

    @cached_property
    def _regressor(self):
        """
        Find the face on which each landmarks sits. Then describe its
        position as a linear combination of the three vertices of that face.

        Represent this as a sparse matrix with a column for each coordinate
        of the source vertices and a row for each coordinate of the
        landmarks.

        Pushing target vertices through the matrix transfers the original
        landmarks to the target mesh.
        """
        import numpy as np
        from scipy.sparse import csc_matrix
        from polliwog.tri import barycentric_coordinates_of_points
        from proximity import faces_nearest_to_points

        landmark_coords = np.array(list(self.landmarks.values()))
        face_indices = faces_nearest_to_points(
            self.source_mesh.v, self.source_mesh.f, landmark_coords
        )
        vertex_indices = self.source_mesh.f[face_indices]
        vertex_coeffs = barycentric_coordinates_of_points(
            self.source_mesh.v[vertex_indices], landmark_coords
        )

        # Note the `.transpose()` at the end. The matrix is initially created
        # from data organized along columns of the result, not rows.
        values = np.repeat(vertex_coeffs, 3, axis=0).ravel()
        indices = (
            (3 * vertex_indices).reshape(-1, 1, 3)
            + np.arange(3, dtype=np.uint64).reshape(-1, 1)
        ).ravel()
        indptr = np.arange(len(values) + 1, step=3)
        shape = (3 * len(self.source_mesh.v), 3 * len(landmark_coords))
        return csc_matrix((values, indices, indptr), shape=shape).transpose()

    def transfer_landmarks_onto(self, target):
        """
        Transfer landmarks onto the given target mesh, which must be in the same
        topology as the source mesh.

        Args:
            target (lacecore.Mesh): Target mesh

        Returns:
            dict: A mapping of landmark names to a np.ndarray with shape `3x1`.
        """
        from ..equality import have_same_topology

        if not target.is_tri:
            raise ValueError("Target mesh must be triangulated")

        if not have_same_topology(self.source_mesh, target):
            raise ValueError("Target mesh must have the same topology")

        target_landmark_coords = (self._regressor * target.v.reshape(-1)).reshape(-1, 3)
        return dict(zip(self.landmarks.keys(), target_landmark_coords))



import lacecore
import numpy as np
from vg.compat import v1 as vg


def _maybe_tqdm(iterable, progress):
    if progress:
        from tqdm import tqdm

        return tqdm(iterable)
    else:
        return iterable


def find_correspondence(
    a, b, atol=1e-4, all_must_match=True, ret_unmatched_b=False, progress=True
):
    """
    Given `a[0], a[1], ..., a[k]` and `b[0], b[1], ..., b[j]`, match each element
    of `a` to the corresponding element of `b`.

    If `a` contains elements which do not exist in `b` and
    `all_must_match=False`, return an index of `-1` for the unmatched elements
    in `a`.

    If `a` contains elements which do not exist in `b` and
    `all_must_match=True`, raise an error. With `all_must_match=True`, unless
    either contains duplicates, `b` is a shuffled copy of `a`, and
    `b[find_correspondence(a, b)]` equals `a`.

    Args:
        a (np.arraylike): `kxn` array.
        b (np.arraylike): `jxn` array.
        atol (float): Match tolerance.
        all_must_match (bool): When `True`, `a` and `b` must contain the
            same elements.
        ret_unmatched_b (bool): When `True`, return a tuple which also contains
            the indices of `b` which were not matched.
        progress (bool): When `True`, show a progress bar.

    Return:
        np.ndarray: Indices of `b` with shape `(k,)`.

    Note:
        This relies on a brute-force algorithm.

        For the interpretation of `atol`, see documentation for `np.isclose`.
    """
    if all_must_match and len(a) != len(b):
        raise ValueError("a and b do not contain the same number of elements")

    a_to_b = np.repeat(-1, len(a))
    b_matched = np.zeros(len(b), dtype=np.bool_)

    for a_index, item in _maybe_tqdm(enumerate(a), progress):
        (indices,) = np.nonzero(np.all(np.isclose(b, item, atol=atol), axis=1))
        if len(indices) >= 1:
            if len(indices) > 1:
                closest_index = np.argmin(vg.euclidean_distance(b[indices], item))
                b_index = indices[closest_index]
            else:
                b_index = indices[0]
            b_matched[b_index] = True
            a_to_b[a_index] = b_index
        elif all_must_match:
            raise ValueError(
                "Couldn't find corresponding element in b for item {} in a".format(
                    a_index
                )
            )

    if ret_unmatched_b:
        (unmatched_b,) = np.where(b_matched == 0)
        return a_to_b, unmatched_b
    else:
        return a_to_b


def restore_correspondence(shuffled_mesh, reference_mesh, atol=1e-4, progress=True):
    """
    Given a reference mesh, reorder the vertices of a shuffled copy to restore
    correspondence with the reference mesh. The vertex set of the shuffled
    mesh and reference mesh must be equal within `atol`. Mutate
    `reference_mesh`. Ignore faces but preserves their integrity.

    Args:
        reference_mesh (lacecore.Mesh): A mesh with the vertices in the
            desired order.
        shuffled_mesh (lacecore.Mesh): A mesh with the same vertex set as
            `reference_mesh`.
        progress (bool): When `True`, show a progress bar.

    Returns:
        tuple: The reordered mesh, and an array which maps old vertices in
        `shuffled_mesh` to new.

    Note:
        This was designed to assist in extracting face ordering and groups from a
        `shuffled_mesh` that "work" with `reference_mesh`, such that face
        ordering and groups can be used on shuffled sets of vertices.

        It relies on a brute-force algorithm.
    """
    v_old_to_new = find_correspondence(
        shuffled_mesh.v, reference_mesh.v, atol=atol, progress=progress
    )
    ordering = np.argsort(v_old_to_new)
    new_mesh = lacecore.reindex_vertices(shuffled_mesh, ordering)
    return new_mesh, v_old_to_new



import lacecore
import numpy as np


def shuffle_vertices(mesh, ret_new_ordering=False):
    """
    Shuffle the mesh's vertex ordering, preserving the integrity of the faces.
    The mesh is mutated.

    Args:
        mesh (lacecore.Mesh): A mesh.
        ret_new_ordering (bool): When `True`, return the new vertex ordering.

    Returns:
        object: When `ret_new_ordering` is `True`, return a tuple containing
            the new mesh and the new vertex ordering. When `False`, return only
            the new mesh.
    """
    ordering = np.random.permutation(len(mesh.v))
    new_mesh = lacecore.reindex_vertices(mesh, ordering)
    if ret_new_ordering:
        return new_mesh, ordering
    else:
        return new_mesh


def shuffle_faces(mesh, ret_new_ordering=False):
    """
    Shuffle the mesh's face ordering. The mesh is mutated.

    Args:
        mesh (lacecore.Mesh): A mesh.
        ret_new_ordering (bool): When `True`, return the new face ordering.

    Returns:
        object: When `ret_new_ordering` is `True`, return a tuple containing
            the new mesh and the new vertex ordering. When `False`, return only
            the new mesh.
    """
    ordering = np.random.permutation(len(mesh.f))
    new_mesh = lacecore.reindex_faces(mesh, ordering)
    if ret_new_ordering:
        return new_mesh, ordering
    else:
        return new_mesh



import numpy as np
from .landmarker import Landmarker


class LandmarkCompositor(object):
    """
    A tool for compositing landmarks from several examples in relation to
    a base mesh. Each example is projected onto the base mesh, then the
    points are averaged.

    The tool takes as input:

    - A base mesh
    - Several examples
        - Mesh (in correspondence with the base mesh)
        - xyz coordinates for one or more landmarks

    And will output:

    - The xyz coordinates of the comosite landmark on the base mesh
    """

    def __init__(self, base_mesh, landmark_names):
        self.base_mesh = base_mesh
        self.landmark_names = set(landmark_names)
        self.examples = []

    def add_example(self, mesh, landmarks):
        # By processing one mesh at a time, they don't all need to be loaded
        # into memory.
        if not set(landmarks.keys()).issuperset(self.landmark_names):
            raise ValueError(
                "Expected examples to contain keys {}".format(
                    ", ".join(self.landmark_names)
                )
            )
        landmarker = Landmarker(source_mesh=mesh, landmarks=landmarks)
        transferred = landmarker.transfer_landmarks_onto(self.base_mesh)
        self.examples.append(transferred)

    @property
    def result(self):
        return {
            name: np.average([example[name] for example in self.examples], axis=0)
            for name in self.landmark_names
        }



from polliwog import Plane
from vg.compat import v1 as vg


def find_opposite_vertices(
    vertices,
    plane_of_symmetry,
    all_must_match=False,
    ret_unmatched=False,
    atol=1e-4,
    progress=True,
):
    """
    Given a plane of symmetry and a point cloud, match each vertex to its
    mirror image.

    When `all_must_match=True`, raise an error if, for any vertex, a
    corresponding mirrored vertex can't be found. When
    `all_must_match=False`, return an index of `-1` for any vertex
    with no mirror image.

    Vertices on the plane of symmetry correspond to themselves.

    Args:
        mesh (lacecore.Mesh): A mesh with `k` vertices.
        atol (float): Match tolerance.
        all_must_match (bool): When `True`, `mesh` must contain a mirror-image
            vertex for each of its vertices.
        progress (bool): When `True`, show a progress bar.

    Return:
        np.ndarray: For each of `mesh.v`, the index in `mesh.v` of its mirror
            image, with shape `(k,)`.

    Note:
        This relies on a brute-force algorithm.

        For the interpretation of `atol`, see documentation for `np.isclose`.
    """
    from .restore_correspondence import find_correspondence

    vg.shape.check(locals(), "vertices", (-1, 3))
    if not isinstance(plane_of_symmetry, Plane):
        raise ValueError("Expected a Plane")

    mirrored_vertices = plane_of_symmetry.mirror_point(vertices)
    return find_correspondence(
        vertices,
        mirrored_vertices,
        ret_unmatched_b=ret_unmatched,
        all_must_match=all_must_match,
        atol=atol,
        progress=progress,
    )



from cached_property import cached_property
import lacecore
import numpy as np
from polliwog import Plane
from vg.compat import v1 as vg
from .landmark_compositor import LandmarkCompositor
from .landmarker import Landmarker

DEFAULT_RADIUS = 0.1


class LandmarkCompositeRecipe(object):
    def __init__(self, recipe):
        """
        Example recipe:

        base_mesh: examples/average.obj
        decimals: 2
        landmarks:
          - knee_left
          - knee_right
        symmetrize:
          reference_point: [0.5, 0.0, 0.0]
          normal: [1.0, 0.0, 0.0]
        examples:
          - id: example01
            mesh: examples/example01.obj
            knee_left: [-10.0, 15.0, 4.0]
            knee_right: [10.0, 14.8, 4.1]
          - id: example02
            mesh: examples/example02.obj
            knee_left: [-11.0, 13.0, 3.5]
            knee_right: [12.0, 12.8, 3.4]
        """
        self.base_mesh_path = recipe["base_mesh"]
        self.decimals = recipe["decimals"]
        self.landmark_names = recipe["landmarks"]
        self.examples = recipe["examples"]
        self.symmetrize = recipe.get("symmetrize", None)

    @classmethod
    def load(cls, recipe_path):
        import yaml

        with open(recipe_path, "r") as f:
            recipe_data = yaml.safe_load(f)
        return cls(recipe_data)

    @cached_property
    def base_mesh(self):
        return lacecore.load_obj(self.base_mesh_path, triangulate=True)

    @property
    def _unsided_landmark_names(self):
        return list(
            set(
                [
                    x.replace("_right", "").replace("_left", "")
                    for x in self.landmark_names
                ]
            )
        )

    @cached_property
    def _plane_of_symmetry(self):
        return Plane(
            point_on_plane=np.array(self.symmetrize["reference_point"]),
            unit_normal=vg.normalize(np.array(self.symmetrize["normal"])),
        )

    def _symmetrize_landmarks(self, landmarks):
        from .symmetrize_landmarks import symmetrize_landmarks_using_plane

        result = {}
        for unsided_name in self._unsided_landmark_names:
            sided_names = [
                "{}_{}".format(unsided_name, side) for side in ["left", "right"]
            ]
            symmetrized = symmetrize_landmarks_using_plane(
                self._plane_of_symmetry,
                np.array([landmarks[k] for k in sided_names]),
            )
            for k, v in zip(sided_names, symmetrized):
                result[k] = v
        return result

    @cached_property
    def composite_landmarks(self):
        base_mesh = self.base_mesh
        compositor = LandmarkCompositor(
            base_mesh=base_mesh, landmark_names=self.landmark_names
        )
        for example in self.examples:
            example_mesh = lacecore.load_obj(example["mesh"], triangulate=True)
            landmarks = {k: example[k] for k in self.landmark_names}
            compositor.add_example(mesh=example_mesh, landmarks=landmarks)
        return compositor.result

    @cached_property
    def symmetrized_landmarks(self):
        return {
            k: v.tolist()
            for k, v in self._symmetrize_landmarks(self.composite_landmarks).items()
        }

    @cached_property
    def reprojected_landmarks(self):
        if self.symmetrize is None:
            landmarks = self.composite_landmarks
        else:
            landmarks = self.symmetrized_landmarks

        inverse_landmarker = Landmarker(source_mesh=self.base_mesh, landmarks=landmarks)

        reprojected = {}
        for example in self.examples:
            mesh = lacecore.load_obj(example["mesh"], triangulate=True)
            reprojected[example["id"]] = inverse_landmarker.transfer_landmarks_onto(
                mesh
            )
        return reprojected

    @property
    def original_and_reprojected_landmarks(self):
        result = {}
        for example in self.examples:
            example_id = example["id"]
            reprojected = self.reprojected_landmarks[example_id]
            result[example_id] = {
                k: {
                    "original": example[k],
                    "reprojected": reprojected[k].tolist(),
                    "displacement": np.round(
                        reprojected[k] - np.array(example[k]), decimals=self.decimals
                    ).tolist(),
                    "euclidean_distance": float(
                        round(
                            vg.euclidean_distance(np.array(example[k]), reprojected[k]),
                            ndigits=self.decimals,
                        )
                    ),
                }
                for i, k in enumerate(self.landmark_names)
            }
        return result

    def to_json(self):
        result = {
            "composited": {k: v.tolist() for k, v in self.composite_landmarks.items()},
            "examples": self.original_and_reprojected_landmarks,
        }
        if self.symmetrize is not None:
            result["composited_and_symmetrized"] = self.symmetrized_landmarks
        return result

    def write_reprojected_landmarks(self, output_dir, radius=DEFAULT_RADIUS):
        import os
        from tri_again import Scene

        original_and_reprojected_landmarks = self.original_and_reprojected_landmarks

        for example in self.examples:
            example_id = example["id"]
            out = os.path.join(output_dir, "{}.dae".format(example_id))

            Scene(point_radius=radius).add_meshes(
                lacecore.load_obj(example["mesh"], triangulate=True)
            ).add_points(
                *[
                    item["original"]
                    for item in original_and_reprojected_landmarks[example_id].values()
                ],
                color="blue"
            ).add_points(
                *[
                    item["reprojected"]
                    for item in original_and_reprojected_landmarks[example_id].values()
                ],
                color="darkgreen"
            ).write(
                out
            )



import numpy as np
from polliwog import Plane
from vg.compat import v1 as vg
from ..symmetry import find_opposite_vertices


def symmetrize_landmarks_using_plane(plane_of_symmetry, landmark_coords):
    """
    TODO
    """
    vg.shape.check(locals(), "landmark_coords", (2, 3))
    if not isinstance(plane_of_symmetry, Plane):
        raise ValueError("plane_of_symmetry should be a Plane")

    mirrored_landmarks = plane_of_symmetry.mirror_point(landmark_coords)

    return np.average(
        np.vstack([landmark_coords.flatten(), np.flipud(mirrored_landmarks).flatten()]),
        axis=0,
    ).reshape(2, 3)


def symmetrize_landmarks_using_topology(
    mesh, plane_of_symmetry, landmark_coords, atol=1e-4
):
    """
    TODO
    """
    from polliwog.tri import barycentric_coordinates_of_points
    from proximity import faces_nearest_to_points

    vg.shape.check(locals(), "landmark_coords", (2, 3))
    if not isinstance(plane_of_symmetry, Plane):
        raise ValueError("plane_of_symmetry should be a Plane")

    # Compute the barycentric coordinates of each landmark.
    indices_of_nearest_faces = faces_nearest_to_points(mesh.v, mesh.f, landmark_coords)
    vertex_indices = mesh.f[indices_of_nearest_faces]
    vertex_coeffs = barycentric_coordinates_of_points(
        mesh.v[vertex_indices], landmark_coords
    )

    # Find the opposite of each of these vertices, and use that to compute the
    # opposite of each landmark.
    # TODO: Add a parameter to `find_opposite_vertices()` that only computes the
    # points we need (i.e. `mesh.v[vertex_indices.flatten()]``). Then use
    # `all_must_match=True`.
    indices_of_opposite_vs = find_opposite_vertices(
        mesh.v,
        plane_of_symmetry=plane_of_symmetry,
        all_must_match=False,
        atol=atol,
    )[vertex_indices]
    if np.any(indices_of_opposite_vs == -1):
        raise ValueError("Some landmarks are near triangles which are not mirrored")

    # TODO: Add a `barys_to_points()` function in polliwog.
    mirrored_landmarks = (
        (vertex_coeffs.reshape(-1, 1) * mesh.v[indices_of_opposite_vs].reshape(-1, 3))
        .reshape(-1, 3, 3)
        .sum(axis=1)
    )

    return np.average(
        np.vstack([landmark_coords.flatten(), np.flipud(mirrored_landmarks).flatten()]),
        axis=0,
    ).reshape(-1, 3)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['entente', 'entente.landmarks']

package_data = \
{'': ['*']}

install_requires = \
['cached_property',
 'lacecore[obj]>=1.0.0,<2',
 'meshlab-pickedpoints>=2.0.0,<3',
 'numpy<1.19.0',
 'ounce>=1.1.1,<2.0',
 'polliwog>=1.0.0,<2.0',
 'scipy',
 'tqdm',
 'vg>=2.0.0']

extras_require = \
{'cli': ['click>7.0,<9.0', 'PyYAML>=5.1', 'tri-again>=1.0.0,<2.0'],
 'landmarker': ['proximity>=1.0.0,<2']}

setup_kwargs = {
    'name': 'entente',
    'version': '1.0.0',
    'description': 'Polygonal meshes in vertex-wise correspondence',
    'long_description': None,
    'author': 'Paul Melnikow',
    'author_email': 'github@paulmelnikow.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/lace/entente',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7,<3.9',
}


setup(**setup_kwargs)



"""
Utilities related to mesh equality.
"""


def attr_has_same_shape(first_obj, second_obj, attr):
    """
    Given two objects, check if the given arraylike attributes of those
    objects have the same shape. If one object has an attribute value of
    ``None``, the other must too.

    Args:
        first_obj (obj): A object with an arraylike ``attr`` attribute.
        second_obj (obj): Another object with an arraylike ``attr`` attribute.
        attr (str): The name of the attribute to test.

    Returns:
        bool: `True` if attributes are the same shape
    """
    # Support legacy `lace.mesh.Mesh`, where attrs are set to `None` instead
    # of empty arrays.
    first, second = getattr(first_obj, attr), getattr(second_obj, attr)
    if first is None or second is None:
        return first is second
    else:
        return first.shape == second.shape


def attr_is_equal(first_obj, second_obj, attr):
    """
    Given two objects, check if the given arraylike attributes of those
    objects are equal. If one object has an attribute value of ``None``, the
    other must too.

    Args:
        first_obj (obj): A object with an arraylike `attr` attribute.
        second_obj (obj): Another object with an arraylike `attr` attribute.
        attr (str): The name of the attribute to test.

    Returns:
        bool: `True` if attributes are equal
    """
    import numpy as np

    # Avoid comparing None's.
    return attr_has_same_shape(first_obj, second_obj, attr) and np.array_equal(
        getattr(first_obj, attr), getattr(second_obj, attr)
    )


def have_same_topology(first_mesh, second_mesh):
    """
    Given two meshes, check if they have the same vertex count and same faces.
    In other words, check if they have the same topology.

    Args:
        first_mesh (lacecore.Mesh): A mesh.
        second_mesh (lacecore.Mesh): Another mesh.

    Returns:
        bool: `True` if meshes have the same topology
    """
    return attr_has_same_shape(first_mesh, second_mesh, "v") and attr_is_equal(
        first_mesh, second_mesh, "f"
    )
